import os

from pdf2image import convert_from_path

from . import files
from . import _C


DPI = _C.DPI


def get_imgs_from_pdf(filedir,
                      dpi=DPI,
                      ):
    imgs = convert_from_path(filedir, dpi=dpi)
    return imgs


def save_imgs_from_pdf(filedir, save_path,
                       dpi=DPI,
                       fext='png',
                       ):
    filename = os.path.split(filedir)[-1]
    imgs = convert_from_path(filedir, dpi=dpi)
    for k, img in enumerate(imgs):
        new_filename = f'{os.path.splitext(filename)[0]}.{k}.{fext.lower()}' if len(imgs) > 1 else f'{os.path.splitext(filename)[0]}.{fext.lower()}'
        new_filedir = os.path.join(save_path, new_filename)
        files.mkdirs_for_filedir(new_filedir)
        img.save(new_filedir, fext.upper())



import copy
import itertools
import random

import numpy as np


RANDOM_STATE = None
SHUFFLE = True


def _check(lst: list):
    assert isinstance(lst, list)
    assert len(lst) > 0


def _check_not_empy(lst: list):
    _check(lst)
    assert len(lst) > 0


def index_list(lst, idxs):
    return list(np.array(lst)[idxs])


def get_shuffled(lst,
                 shuffle=SHUFFLE,
                 random_state=RANDOM_STATE,
                 ):
    new_lst = copy.deepcopy(lst)
    if shuffle:
        random.seed(random_state)
        random.shuffle(new_lst)
    return new_lst


def get_shared_shuffled(*args,
                        shuffle=SHUFFLE,
                        random_state=RANDOM_STATE,
                        ):
    idxs = list(range(0, len(args[0])))
    new_idxs = get_shuffled(idxs,
                            random_state=random_state,
                            shuffle=shuffle,
                            )
    new_args = []
    for lst in args:
        new_args += [copy.deepcopy(index_list(lst, new_idxs))]
    return new_args


def check_same_class(elements):
    return all([type(e) == type(elements[0]) for e in elements])


def get_max_elements(elements):
    assert check_same_class(elements), 'all objects must be of the same class'
    max_elements = []
    max_e = max(elements)
    for e in elements:
        if e >= max_e:
            max_elements += [e]
    return [True if e in max_elements else False for e in elements]


def get_min_elements(elements):
    assert check_same_class(elements), 'all objects must be of the same class'
    min_elements = []
    min_e = min(elements)
    for e in elements:
        if e <= min_e:
            min_elements += [e]
    return [True if e in min_elements else False for e in elements]


def list_product(*args):
    return list(itertools.product(*args))  # just a wrap


def split_list_in_batches(lst, batch_size):
    batches = []
    index = 0
    while index < len(lst):
        batches.append(lst[index: index + batch_size])
        index += batch_size
    return batches


def flat_list(list_of_lists):
    return sum(list_of_lists, [])


def get_random_item(lst):
    _check_not_empy(lst)
    idx = 0 if len(lst) == 1 else random.randint(0, len(lst) - 1)
    return lst[idx]


def get_bootstrap(lst: list, n,
                  random_state=RANDOM_STATE,
                  replace=True,
                  ):
    if replace:
        random.seed(random_state)
        bootstrap = [get_random_item(lst) for _ in range(0, n)]  # faster than numpy.choice
        return bootstrap
    else:
        np.random.seed(random_state)
        bootstrap = np.choice(lst, size=n, replace=False)
        return bootstrap


def merge_lists(*args):
    merged = list(itertools.chain(*args))
    return merged


def delete_from_list(lst: list, elements_to_remove: list):
    removed_elements = []
    new_lst = []
    for e in lst:
        if e in elements_to_remove:
            removed_elements += [e]
        else:
            new_lst += [e]
    return new_lst, removed_elements


def all_elements_are_equal(lst: list):
    return lst.count(lst[0]) == len(lst)


def get_lists_intersection(lst1, lst2):
    intersection = list(set(lst1).intersection(set(lst2)))
    return intersection


def check_lists_are_different(lst1, lst2):
    for x in lst1:
        if x in lst2:
            return False
    return True


def check_lists_are_equal(lst1, lst2,
                          checks_len=True,
                          ):
    c = set(lst1) == set(lst2)
    if checks_len:
        c = c and len(lst1) == len(lst2)
    return c



from . import lists
from nested_dict import nested_dict


def get_random_key(d: dict):
    keys = list(d.keys())
    key = lists.get_random_item(keys)
    return key


def update_dicts(dicts):
    d = {}
    for _d in dicts:
        assert isinstance(_d, dict)
        d.update(_d)
    return d


def along_dict_obj_method(obj_dict, obj_method,
                          obj_args=[],
                          obj_kwargs={},
                          ):
    _obj_dict = nested_dict(obj_dict)
    for keys_as_tuple, value in _obj_dict.items_flat():
        getattr(value, obj_method)(*obj_args, **obj_kwargs)
    return _obj_dict.to_dict()


import setuptools


with open('version', 'r') as fh:
    version = fh.read()
with open('README.md', 'r') as fh:
    long_description = fh.read()
with open('requirements.txt') as f:
    required = f.read().splitlines()
classifiers = [
    'Programming Language :: Python :: 3',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
]
setuptools.setup(name='cambriantools',
                 version=version,
                 description='generic description',
                 python_requires='>=3.10',
                 install_requires=required,
                 keywords='experimental',
                 author='Oscar Pimentel Fuentes',
                 author_email='oscarlo.pimentel@gmail.com',
                 #  include_package_data=True,
                 packages=setuptools.find_packages(),
                 long_description=long_description,
                 long_description_content_type='text/markdown',
                 license='MIT licence',
                 classifiers=classifiers,
                 )



import os
import pickle

from . import _C


VERBOSE = _C.VERBOSE


def get_filedirs(rootdir: str,
                 fext=None,
                 ):
    """
    Get a list of filedirs in all subdirs with extention .fext

    Parameters:
    -----------
    rootdir: start path to search
    fext: file extention. None: search for all extentions

    Return:
    -------
    filedirs (list[srt]): list of filedirs
    """
    filedirs = []
    for root, dirs, files in os.walk(rootdir):
        for f in files:
            if fext is None or f.split('.')[-1]==fext: # dont add if none
                filedirs += [f'{root}/{f}']
    return filedirs


def mkdirs_for_filedir(filedir):
    filepath = os.path.split(filedir)[0]
    os.makedirs(filepath, exist_ok=True)


def save_txt(filedir, string):
    text_file = open(filedir, 'w')
    text_file.write(string)
    text_file.close()


def save_pickle(filedir: str, file: object):
    if not type(filedir) == str:
        raise TypeError('wrong filedir type')
    mkdirs_for_filedir(filedir)
    file_pi = open(filedir, 'wb')
    pickle.dump(file, file_pi)
    file_pi.close()


def load_pickle(filedir: str,
                return_none_if_missing=False,
                verbose=VERBOSE,
                ):
    """
    Parameters:
    -----------
    filedir:
        The filedir of file to read

    Returns:
    --------
    file (object):
        The read object from disk
    """
    if filedir is None:
        return None

    if not filedir_exists(filedir):
        if return_none_if_missing:
            return None
        else:
            raise Exception(f'no file in {filedir}')

    if verbose==1:
        prints.print_blue(f'> loading: {filedir}')

    pickle_in = open(filedir,'rb')
    file = pickle.load(pickle_in)
    pickle_in.close()
    return file

"""

# import time
# from . import strings
# from . import prints
# from . import lists
# from nested_dict import nested_dict
# from .times import get_date_hour
# from shutil import copyfile

# KEY_KEY_SEP_CHAR = _C.KEY_KEY_SEP_CHAR
# KEY_VALUE_SEP_CHAR = _C.KEY_VALUE_SEP_CHAR
# KFOLF_CHAR = _C.KFOLF_CHAR
# FILEDIR = '_filedir' # ../../save/data.txt
# ROOTDIR = '_rootdir' # ../../save/
# FILENAME = '_filename' # data.txt
# CFILENAME = '_cfilename' # data
# FEXT = '_fext' # txt
# FILESIZE_FACTOR = 1e-6 # in mbs
# IMBALANCED_KF_MODE = 'error' # error ignore oversampling
# RANDOM_STATE = None

###################################################################################################################################################

def filedir_exists(filedir:str):
    return os.path.isfile(filedir)

def path_exists(rootdir:str):
    return os.path.isdir(rootdir)

def get_filesize(filedir:str):
    return os.path.getsize(filedir)*FILESIZE_FACTOR if filedir_exists(filedir) else None # in mb

###################################################################################################################################################

def delete_filedirs(filedirs:list,
    verbose=VERBOSE,
    ):
    all_success = sum([delete_filedir(filedir, verbose) for filedir in filedirs])>0
    return all_success

def delete_filedir(filedir:str,
    verbose=VERBOSE,
    ):
    success = False
    if not filedir is None:
        if filedir_exists(filedir):
            if verbose==1:
                prints.print_red(f'> deleting: {filedir}')
            os.remove(filedir)
            success = True

    return success

def copy_filedir(src_filedir, dst_filedir):
    if src_filedir==dst_filedir:
        return False
    copyfile(src_filedir, dst_filedir)
    return True

###################################################################################################################################################

def create_dir(new_dir:str,
    iterative:bool=True,
    verbose:int=VERBOSE,
    ):
    if path_exists(new_dir):
        return

    if verbose==1:
        prints.print_yellow(f'> creating dir: {new_dir}')

    if iterative:
        create_dir_iterative(new_dir, verbose=int(verbose==2))
    else:
        create_dir_individual(new_dir, verbose=int(verbose==2))
    return


def create_dir_individual(new_dir:str,
    verbose:int=VERBOSE,
    ):
    if path_exists(new_dir):
        return

    if verbose==1:
        prints.print_yellow(f'>> creating dir: {new_dir}')
    os.mkdir(new_dir)

def create_dir_iterative(new_dir:str,
    verbose:int=VERBOSE,
    ):
    if path_exists(new_dir):
        return

    dirs = new_dir.split('/')
    new_dir = ''
    for f in dirs:
        new_dir += f+'/'
        create_dir_individual(new_dir, verbose=verbose)

###################################################################################################################################################

def save_pickle(filedir:str, file:object,
    verbose=VERBOSE,
    ):
    '''
    Parameters
    ----------
    filedir: filedir of file to save
    file: object to save. Be careful with cuda serialized objects
    '''
    assert isinstance(filedir, str)
    if verbose==1:
        prints.print_green(f'> saving: {filedir}')

    create_dir('/'.join(filedir.split('/')[:-1]), verbose=int(verbose==2))
    file_pi = open(filedir, 'wb')
    pickle.dump(file, file_pi)
    file_pi.close()


###################################################################################################################################################

def save_time_stamp(rootdir,
    extra_info={},
    cfilename='time_stamp',
    ):
    create_dir(rootdir)
    filedir = f'{rootdir}/{cfilename}.txt'
    with open(filedir, 'w') as text_file:
        date, hour = get_date_hour()
        text_file.write(f'{date}\n')
        text_file.write(f'{hour}\n')
        if len(extra_info.keys())>0:
            text_file.write(f'\n')
            text_file.write(f'[extra_info]\n')
            for k in extra_info.keys():
                text_file.write(f'{k}={extra_info[k]}\n')

###################################################################################################################################################

def get_dict_from_filedir(filedir:str,
    key_key_separator:str=KEY_KEY_SEP_CHAR,
    key_value_separator:str=KEY_VALUE_SEP_CHAR,
    ):
    splits = filedir.split('/')
    ret_dict = {
        FILEDIR:filedir,
        ROOTDIR:'/'.join(splits[:-1]),
        FILENAME:'.'.join(splits[-1].split('.')),
        CFILENAME:'.'.join(splits[-1].split('.')[:-1]),
        FEXT:splits[-1].split('.')[-1],
    }
    ret_dict.update(strings.get_dict_from_string(ret_dict[CFILENAME],
        key_key_separator,
        key_value_separator,
    ))
    return ret_dict

def search_for_filedirs(rootdir:str,
    string_query:list=[''],
    string_filter:list=[],
    fext:str=None,
    verbose:int=VERBOSE,
    sort:bool=False,
    ):
    '''
    Get a list of filedirs in all subdirs with extention .fext.
    Also, uses filters of key strings.

    Parameters
    ----------
    rootdir (srt): start path to search
    string_query (list[srt]): (optional) list with string queries that have to appear in all the cfilenames.
    string_filter(list[str]): (optional) list with string queries that don't have to appear in all the cfilenames.
    fext (srt): (optional) file extention. Default is None: search for all extentions
    verbose (int): verbosity of method

    Return
    ----------
    filesret (list[srt]): list of filedirs that meet the conditions
    '''
    prints.print_bar()
    filedirs = get_filedirs(rootdir, fext=fext)
    print(f'found filedirs: ({rootdir})')
    for k,filedir in enumerate(filedirs):
        filesize = get_filesize(filedir)
        print(f'({k}) - {filedir} - {filesize:.3f}[mbs]')

    if sort:
        filedirs.sort(key=str.lower)

    filedirs_res = []
    for filedir in filedirs:
        filedict = get_dict_from_filedir(filedir)
        cfilename = filedict[CFILENAME]
        if strings.query_strings_in_string(string_query, cfilename) and not strings.query_strings_in_string(string_filter, cfilename):
            filedirs_res.append(filedir)

    prints.print_bar()
    print(f'filedirs after searching with filters: ({rootdir})')
    for k,filedir in enumerate(filedirs_res):
        filesize = get_filesize(filedir)
        print(f'({k}) - {filedir} - {filesize:.3f}[mbs]')
    prints.print_bar()
    return filedirs_res

def get_filedirs(rootdir:str,
    fext:str=None,
    ):
    '''
    Get a list of filedirs in all subdirs with extention .fext

    Parameters
    ----------
    rootdir: start path to search
    fext: file extention. None: search for all extentions

    Return
    ----------
    filedirs (list[srt]): list of filedirs
    '''
    filedirs = []
    for root, dirs, files in os.walk(rootdir):
        level = root.replace(rootdir, '').count(os.sep)
        indent = ' ' * 4 * (level)+'> '
        subindent = ' ' * 4 * (level + 1)+'- '
        for f in files:
            if fext is None or f.split('.')[-1]==fext: # dont add if none
                filedirs += [f'{root}/{f}']
    return filedirs

def get_roodirs(rootdir):
    depth = len(rootdir.split('/'))
    roodirs = []
    for root, dirs, files in os.walk(rootdir):
        sub_depth = len(root.split('/'))
        if sub_depth-depth==1:
            roodirs += [root]
    return roodirs

def get_nof_filedirs(filedir:str,
    fext:str=None,
    ):
    '''
    return the count of filenames with an extention .fext
    '''
    return len(get_filedirs(filedir, fext=fext))

def print_all_filedirs(filedir):
    print(f'total files in {filedir}: {get_nof_filedirs(filedir)}')
    for root, dirs, files in os.walk(filedir):
        level = root.replace(filedir, '').count(os.sep)
        indent = ' ' * 4 * (level)+'> '
        print(f'{indent}{os.path.basename(root)}/')
        subindent = ' ' * 4 * (level + 1)+'- '
        for f in files:
            print(f'{subindent}{f}')

###################################################################################################################################################

def getmtime(filedir):
    return time.ctime(os.path.getmtime(filedir))

def getctime(filedir):
    return time.ctime(os.path.getmtime(filedir))

def get_newest_filedir(filedirs,
    mode='m',
    ):
    if mode=='c':
        dates = [getctime(f) for f in filedirs]
    elif mode=='m':
        dates = [getmtime(f) for f in filedirs]
    else:
        raise Exception(f'{mode}')
    max_date = max(dates)
    filedir = filedirs[dates.index(max_date)]
    return filedir

###################################################################################################################################################

class PFile(object):
    def __init__(self, filedir,
        file=None,
        verbose=VERBOSE,
        ):
        self.set_filedir(filedir)
        self.set_file(file)
        self.verbose = verbose
        self.reset()

    def set_filedir(self, filedir):
        self.is_dummy = filedir is None
        self.filedir = filedir

    def set_file(self, file):
        self.file = file

    def reset(self):
        if not self.is_dummy:
            self.filedict = get_dict_from_filedir(self.filedir)
            self.rootdir = self.filedict['_rootdir']
            self.filename = self.filedict['_filename']
            self.cfilename = self.filedict['_cfilename']
            self.fext = self.filedict['_fext']
            self.disk_size = get_filesize(self.filedir)
            self.last_state = 'idle'

            ### attemp reading
            if self.file is None:
                if not self.disk_size is None: # file exists in disk
                    self.load()
                else:
                    raise Exception(f'no PFile.file and {self.filedir} does not exist in disk')
            else:
                pass
            return

    def __getitem__(self, idx):
        return self.filedict[idx]

    def __repr__(self):
        if not self.is_dummy:
            d = {
                'cfilename':self.cfilename,
                'fext':self.fext,
                'rootdir':self.rootdir,
                'file_class':type(self.file).__name__,
                'disk_size':f'{self.disk_size:.3f} [mb]' if not self.disk_size is None else None,
                'last_state':self.last_state,
                }
            txt = f'PFile({strings.get_string_from_dict(d, key_key_separator=", ")})'
        else:
            txt = f'PFile(dummy)'
        return txt

    def get_file(self):
        return self.file

    def __call__(self):
        return self.get_file()

    def save(self,
        copy_filedirs=[],
        ):
        if not self.is_dummy:
            return self._save(
                copy_filedirs,
                )
        else:
            return

    def _save(self,
        copy_filedirs=[],
        ):
        filedirs = [self.filedir]+copy_filedirs
        for filedir in filedirs:
            save_pickle(filedir, self.file)
        self.last_state = 'saved'
        return

    def load(self):
        if not self.is_dummy:
            return self._load()

    def _load(self):
        file = load_pickle(self.filedir)
        self.set_file(file)
        self.reset()
        self.last_state = 'loaded'
        return

###################################################################################################################################################

def gather_files(rootdir,
    fext:str=None,
    ):
    filedirs = get_filedirs(rootdir, fext=fext)
    return [PFile(filedir) for filedir in filedirs]

def gather_files_by_id(rootdir,
    id_key='id',
    key_key_separator:str=KEY_KEY_SEP_CHAR,
    key_value_separator:str=KEY_VALUE_SEP_CHAR,
    fext:str=None,
    ):
    if id_key is None:
        files = gather_files(rootdir,
            fext,
            )
        file_ids = [f.cfilename for f in files]
        return files, file_ids

    else:
        filedirs = get_filedirs(rootdir, fext=fext)
        files = []
        file_ids = []
        for filedir in filedirs:
            filedict = get_dict_from_filedir(filedir,
                key_key_separator,
                key_value_separator,
                )
            files += [PFile(filedir)]
            file_ids += [filedict[id_key]]
        return files, file_ids

def get_kfold_rootdirs_dict(rootdir,
    kf_str=KFOLF_CHAR,
    ):
    '''
    {
    [kf_set][kf] = rootdir
    }
    '''
    rootdirs = get_roodirs(rootdir)
    kfold_rootdirs_dict = nested_dict()
    for rd in rootdirs:
        kf, kf_set = rd.split('/')[-1].split(kf_str)
        kfold_rootdirs_dict[kf_set][kf] = rd
    return kfold_rootdirs_dict.to_dict()

def gather_files_by_kfold(rootdir, _kf, kf_set,
    id_key=None,
    key_key_separator:str=KEY_KEY_SEP_CHAR,
    key_value_separator:str=KEY_VALUE_SEP_CHAR,
    fext:str=None,
    kf_str=KFOLF_CHAR,
    imbalanced_kf_mode=IMBALANCED_KF_MODE,
    random_state=RANDOM_STATE,
    kfs=None,
    returns_all_kf_files=False,
    ):
    '''
    format must be .../kf@kf_set/files

    imbalanced_kf_mode is used just in case a k-fold iteration is missing and there is an imbalance on files (mostly for debuggin while results are still iterating)
    the normal case is when all k-fold has the same amount of files!
    '''
    kfold_rootdirs_dict = get_kfold_rootdirs_dict(rootdir,
        kf_str=kf_str,
    )
    kf = str(_kf)
    if kf=='.': # gather files from all kf values
        all_kf_files = {}
        all_kf_file_ids = {}
        kfs = list(kfold_rootdirs_dict[kf_set].keys()) if kfs is None else kfs
        for _kf in kfs:
            kfrd = kfold_rootdirs_dict[kf_set][_kf]
            _files, _file_ids = gather_files_by_id(kfrd,
                id_key,
                key_key_separator,
                key_value_separator,
                fext,
                )
            assert len(_files)>0, f'no files for kf={_kf}'
            all_kf_files[_kf] = _files
            all_kf_file_ids[_kf] = [f'{_kf}{kf_str}{_fid}' for _fid in _file_ids]

        # print(all_kf_files); print(all_kf_file_ids)
        max_len = max([len(all_kf_files[_kf]) for _kf in kfs])
        if imbalanced_kf_mode=='ignore':
            pass

        elif imbalanced_kf_mode=='error':
            for _kf in kfs:
                assert len(all_kf_files[_kf])==max_len, f'not equal size of kf files in all kf values {[len(all_kf_files[_kf]) for _kf in kfs]}'

        elif imbalanced_kf_mode=='oversampling':
            for _kf in kfs:
                _len = len(all_kf_files[_kf])
                k_repeat = max_len-_len
                if k_repeat>0:
                    idxs = list(range(0, _len))
                    new_idxs = lists.get_bootstrap(idxs, k_repeat,
                        random_state=random_state,
                        )
                    all_kf_files[_kf] += [all_kf_files[_kf][idx] for idx in new_idxs]
                    all_kf_file_ids[_kf] += [all_kf_file_ids[_kf][idx] for idx in new_idxs]
        else:
            raise Exception(f'imbalanced_kf_mode={imbalanced_kf_mode}')

        # print(all_kf_files); print(all_kf_file_ids)
        files = []
        file_ids = []
        for _kf in kfs:
            files += all_kf_files[_kf]
            file_ids += all_kf_file_ids[_kf]

        if returns_all_kf_files:
            return files, file_ids, kfs, all_kf_files
        else:
            return files, file_ids, kfs

    else: # gather files from an specific kf value
        if not kf_set in kfold_rootdirs_dict.keys():
            return [], [], []
        kfold_rootdirs_dict_set = kfold_rootdirs_dict[kf_set]
        if not kf in kfold_rootdirs_dict_set.keys():
            return [], [], []
        kfrd = kfold_rootdirs_dict_set[str(kf)]
        files, file_ids = gather_files_by_id(kfrd,
            id_key,
            key_key_separator,
            key_value_separator,
            fext,
            )
        return files, file_ids, [kf]
"""



import os
import copy

import numpy as np
from termcolor import colored


N_DECIMALS = 3
BAR_SIZE = None
JUPYTER_NOTEBOOK_BAR_SIZE = 100
REMOVE_ZERO = True
ALPHABET_CHARS = 'abcdefghijklmnopqrstuvwxyz'
NUMBER_CHARS = '0123456789'
MIDDLE_LINE_CHAR = '─'
BAR_LAST_CHAR = '>'
BOT_SQUARE_CHAR = '▄'
TOP_SQUARE_CHAR = '▀'
BAR_CHAR = '-'
DOBLE_BAR_CHAR = '═'
KFOLF_CHAR = '@'
KEY_KEY_SEP_CHAR = '~'
KEY_VALUE_SEP_CHAR = '='
NAN_CHAR = '─'
PM_CHAR = '±'


def get_nof_chars(string, chars):
    count = 0
    for char in chars:
        count += string.count(char)
    return count


def get_nof_alphabet_chars(string,
                           alphabet_chars=ALPHABET_CHARS,
                           uses_lower=True,
                           ):
    new_string = string.lower() if uses_lower else string
    return get_nof_chars(new_string, alphabet_chars)


def get_nof_number_chars(string,
                         number_chars=NUMBER_CHARS,
                         ):
    return get_nof_chars(string, number_chars)


def clean_string_by_chars(string, chars):
    new_string = copy.deepcopy(string)
    count_d = {}
    for char in chars:
        count = new_string.count(char)
        count_d[f'{char}'] = count
        new_string = new_string.replace(char, '')
    return new_string, count_d


def get_raw_numpy_repr(x,
                       name='x',
                       ):
    txt = f'{name} = {x.tolist()}'
    return txt


def get_string_from_dict(d: dict,
                         key_key_separator=KEY_KEY_SEP_CHAR,
                         key_value_separator=KEY_VALUE_SEP_CHAR,
                         keeps_none=True,
                         ):
    ret = key_key_separator.join([f'{key}{key_value_separator}{d[key]}' for key in d.keys() if keeps_none or not d[key] is None])
    return ret


def get_formated_method(method_name, method_atributes_dict,
                        inter_str=' ',
                        ):
    txt = get_string_from_dict(method_atributes_dict, '; ', '=',
                               keeps_none=False,
                               )
    txt = f'{method_name}{inter_str}(' + txt + ')'
    return txt


def get_dict_from_string(string,
                         key_key_separator=KEY_KEY_SEP_CHAR,
                         key_value_separator=KEY_VALUE_SEP_CHAR,
                         ):
    pairs = string.split(key_key_separator)
    ret = {}
    for p in pairs:
        split = p.split(key_value_separator)
        if len(split) == 2:
            key, value = split
            ret[key] = value
    return ret


def string_replacement(string, replace_dict):
    """
    Reeplace multiple sub-strings in a string.
    Example:
    string: Hi to all the hidden demons.
    replace_dict: {'hi':'Hellow', 'demons': 'dogs', ' ':'_'}
    Return: Hellow_to_all_the_hidden_dogs.

    Parameters
    ----------
    string: str
        The target string
    replace_dict: dict
        This is a dictionary where the keys are the target sub-string to be replaced and the values are the sub-string replacement. The replacement are in the key order.

    Returns
    -------
    new_string: str
        The new string
    """
    assert type(replace_dict) == dict
    new_string = copy(string)
    for key in replace_dict:
        new_string = new_string.replace(key, replace_dict[key])
    return new_string


def query_strings_in_string(query_strings: list, string: str,
                            mode='or',
                            ):
    """
    Search if at least one query_string is in a string
    """
    assert type(query_strings) == list
    assert type(string) == str
    values = [int(i in string) for i in query_strings]
    if mode == 'or':
        return sum(values) > 0
    elif mode == 'and':
        return sum(values) == len(values)
    else:
        raise Exception(f'{mode}')


def delete_string_chars(s, chars):
    return ''.join([c for c in s if c not in chars])


def random_repeat_string(s, a, b):
    r = np.random.randint(a, b)
    return s * r


def random_augment_string(s, a, b):
    return ''.join([random_repeat_string(c, a, b) for c in s])


def _format_float(x,
                  n_decimals=N_DECIMALS,
                  remove_zero=REMOVE_ZERO,
                  ):
    if np.isnan(x):
        return NAN_CHAR
    txt = format(x, f',.{n_decimals}f')
    if remove_zero and abs(x) < 1:
        txt = txt.replace('0.', '.')
    return txt


def _format_int(x):
    if np.isnan(x):
        return NAN_CHAR
    return f'{x:,}'


def xstr(x,
         n_decimals=N_DECIMALS,
         remove_zero=REMOVE_ZERO,
         add_pos=False,
         ):
    pchar = '+' if add_pos and x > 0 else ''
    if x is None:
        return NAN_CHAR
    elif type(x) == str:
        return x
    elif type(x) == int:
        return pchar + _format_int(x)  # int
    elif type(x) == float:
        return pchar + _format_float(x, n_decimals, remove_zero)  # float
    elif type(x) == np.int64:
        return pchar + _format_int(x)  # int
    elif type(x) == np.float64:
        return pchar + _format_float(x, n_decimals, remove_zero)  # float
    else:
        raise Exception(f'{type(x)}')


def latex_sub_superscript(s,
                          subscript=' ',
                          superscript=' ',
                          ):
    return s + '${}_{' + subscript + '}^{' + superscript + '}$'


def latex_bf(s):
    if s is None:
        return ''
    else:
        return '$\\bf{(' + str(s) + ')}$'


def latex_bf_alphabet_count(k,
                            extra_string=None,
                            string_length=1,
                            adds_space=True,
                            ):
    if k is None:
        return ''
    c = alphabet_count(k, string_length)
    s = '' if extra_string is None else f'.{extra_string}'
    txt = latex_bf(f'{c}{s}')
    txt = txt + ' ' if adds_space else txt
    return txt


def alphabet_count(k,
                   string_length=None,
                   base=ALPHABET_CHARS,
                   ):
    if k is None:
        return None
    assert k >= 0
    base_first = base[0]
    string_length = len(base) if string_length is None else string_length
    assert k < len(base)**string_length
    res = ""
    b = len(base)
    while k:
        res += base[k % b]
        k //= b
    txt = res[::-1] or base_first
    txt = base_first * (string_length - len(txt)) + txt
    return txt


def get_bar(char=MIDDLE_LINE_CHAR,
            N=BAR_SIZE,
            init_string='',
            ):
    if N is None:
        try:
            N = os.get_terminal_size().columns
        except OSError:
            N = JUPYTER_NOTEBOOK_BAR_SIZE
    return init_string + char * N


def color_str(txt, color):
    if color is None or txt == '':
        return txt
    else:
        return colored(txt, color)
